@@ -2,8 +2,8 @@
 
   class.c -
 
-  $Author: ocean $
-  $Date: 2004/07/15 11:42:46 $
+  $Author: nobu $
+  $Date: 2005/05/16 13:28:59 $
   created at: Tue Aug 10 15:05:44 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -92,6 +92,9 @@ rb_class_init_copy(clone, orig)
     if (RCLASS(clone)->super != 0) {
 	rb_raise(rb_eTypeError, "already initialized class");
     }
+    if (FL_TEST(orig, FL_SINGLETON)) {
+	rb_raise(rb_eTypeError, "can't copy singleton class");
+    }
     return rb_mod_init_copy(clone, orig);
 }
 
