@@ -2,8 +2,8 @@
 
   dir.c -
 
-  $Author: knu $
-  $Date: 2004/07/17 18:24:15 $
+  $Author: matz $
+  $Date: 2005/07/27 02:10:56 $
   created at: Wed Jan  5 09:51:01 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -63,8 +63,8 @@ char *strchr _((char*,char));
 
 #include "util.h"
 
-#ifndef HAVE_LSTAT
-#define lstat(path,st) stat(path,st)
+#if !defined HAVE_LSTAT && !defined lstat
+#define lstat stat
 #endif
 
 #define FNM_NOESCAPE	0x01
@@ -236,7 +236,11 @@ static void
 free_dir(dir)
     struct dir_data *dir;
 {
-    if (dir && dir->dir) closedir(dir->dir);
+    if (dir) {
+	if (dir->dir) closedir(dir->dir);
+	if (dir->path) free(dir->path);
+    }
+    free(dir);
 }
 
 static VALUE dir_close _((VALUE));
@@ -299,6 +303,7 @@ dir_initialize(dir, dirname)
  *  the block, and <code>Dir::open</code> returns the value of the
  *  block.
  */
+
 static VALUE
 dir_s_open(klass, dirname)
     VALUE klass, dirname;
@@ -404,6 +409,7 @@ dir_each(dir)
     struct dirent *dp;
 
     GetDIR(dir, dirp);
+    rewinddir(dirp->dir);
     for (dp = readdir(dirp->dir); dp != NULL; dp = readdir(dirp->dir)) {
 	rb_yield(rb_tainted_str_new(dp->d_name, NAMLEN(dp)));
 	if (dirp->dir == NULL) dir_closed();
@@ -459,10 +465,11 @@ dir_seek(dir, pos)
     VALUE dir, pos;
 {
     struct dir_data *dirp;
+    off_t p = NUM2OFFT(pos);
 
-#ifdef HAVE_SEEKDIR
     GetDIR(dir, dirp);
-    seekdir(dirp->dir, NUM2INT(pos));
+#ifdef HAVE_SEEKDIR
+    seekdir(dirp->dir, p);
     return dir;
 #else
     rb_notimplement();
@@ -538,24 +545,42 @@ dir_close(dir)
 
 static void
 dir_chdir(path)
-    const char *path;
+    VALUE path;
 {
-    if (chdir(path) < 0)
-	rb_sys_fail(path);
+    if (chdir(RSTRING(path)->ptr) < 0)
+	rb_sys_fail(RSTRING(path)->ptr);
 }
 
 static int chdir_blocking = 0;
 static VALUE chdir_thread = Qnil;
 
+struct chdir_data {
+    VALUE old_path, new_path;
+    int done;
+};
+
 static VALUE
-chdir_restore(path)
-    char *path;
+chdir_yield(args)
+    struct chdir_data *args;
 {
-    chdir_blocking--;
-    if (chdir_blocking == 0)
-	chdir_thread = Qnil;
-    dir_chdir(path);
-    free(path);
+    dir_chdir(args->new_path);
+    args->done = Qtrue;
+    chdir_blocking++;
+    if (chdir_thread == Qnil)
+	chdir_thread = rb_thread_current();
+    return rb_yield(args->new_path);
+}
+
+static VALUE
+chdir_restore(args)
+    struct chdir_data *args;
+{
+    if (args->done) {
+	chdir_blocking--;
+	if (chdir_blocking == 0)
+	    chdir_thread = Qnil;
+	dir_chdir(args->old_path);
+    }
     return Qnil;
 }
 
@@ -605,19 +630,18 @@ dir_s_chdir(argc, argv, obj)
     VALUE obj;
 {
     VALUE path = Qnil;
-    char *dist = "";
 
     rb_secure(2);
     if (rb_scan_args(argc, argv, "01", &path) == 1) {
 	SafeStringValue(path);
-	dist = RSTRING(path)->ptr;
     }
     else {
-	dist = getenv("HOME");
+	const char *dist = getenv("HOME");
 	if (!dist) {
 	    dist = getenv("LOGDIR");
 	    if (!dist) rb_raise(rb_eArgError, "HOME/LOGDIR not set");
 	}
+	path = rb_str_new2(dist);
     }
 
     if (chdir_blocking > 0) {
@@ -626,14 +650,15 @@ dir_s_chdir(argc, argv, obj)
     }
 
     if (rb_block_given_p()) {
+	struct chdir_data args;
 	char *cwd = my_getcwd();
-	chdir_blocking++;
-	if (chdir_thread == Qnil)
-	    chdir_thread = rb_thread_current();
-	dir_chdir(dist);
-	return rb_ensure(rb_yield, path, chdir_restore, (VALUE)cwd);
+
+	args.old_path = rb_tainted_str_new2(cwd); free(cwd);
+	args.new_path = path;
+	args.done = Qfalse;
+	return rb_ensure(chdir_yield, (VALUE)&args, chdir_restore, (VALUE)&args);
     }
-    dir_chdir(dist);
+    dir_chdir(path);
 
     return INT2FIX(0);
 }
@@ -735,13 +760,8 @@ dir_s_mkdir(argc, argv, obj)
     }
 
     check_dirname(&path);
-#ifndef _WIN32
     if (mkdir(RSTRING(path)->ptr, mode) == -1)
 	rb_sys_fail(RSTRING(path)->ptr);
-#else
-    if (mkdir(RSTRING(path)->ptr) == -1)
-	rb_sys_fail(RSTRING(path)->ptr);
-#endif
 
     return INT2FIX(0);
 }
@@ -870,6 +890,7 @@ glob_func_caller(val)
     VALUE val;
 {
     struct glob_args *args = (struct glob_args *)val;
+
     (*args->func)(args->c, args->v);
     return Qnil;
 }
@@ -891,42 +912,60 @@ glob_call_func(func, path, arg)
     return status;
 }
 
+static int glob_helper _((const char *path, const char *sub, int flags, void (*func)(const char *,VALUE), VALUE arg));
+
 static int
 glob_helper(path, sub, flags, func, arg)
-    char *path;
-    char *sub;
+    const char *path;
+    const char *sub;
     int flags;
-    void (*func) _((const char*, VALUE));
+    void (*func) _((const char *, VALUE));
     VALUE arg;
 {
     struct stat st;
-    char *p, *m;
+    const char *p, *m;
     int status = 0;
+    char *buf = 0;
+    char *newpath = 0;
 
     p = sub ? sub : path;
     if (!has_magic(p, 0, flags)) {
+	if (
 #if defined DOSISH
-	remove_backslashes(path);
+	    TRUE
 #else
-	if (!(flags & FNM_NOESCAPE)) remove_backslashes(p);
+	    !(flags & FNM_NOESCAPE)
 #endif
+	    )
+	{
+	    newpath = strdup(path);
+	    if (sub) {
+		p = newpath + (sub - path);
+		remove_backslashes(newpath + (sub - path));
+		sub = p;
+	    }
+	    else {
+		remove_backslashes(newpath);
+		p = path = newpath;
+	    }
+	}
 	if (lstat(path, &st) == 0) {
 	    status = glob_call_func(func, path, arg);
-	    if (status) return status;
 	}
 	else if (errno != ENOENT) {
 	    /* In case stat error is other than ENOENT and
 	       we may want to know what is wrong. */
-	    rb_sys_warning(path);
+	    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)path, 0);
 	}
-	return 0;
+	xfree(newpath);
+	return status;
     }
 
     while (p && !status) {
 	if (*p == '/') p++;
 	m = strchr(p, '/');
 	if (has_magic(p, m, flags)) {
-	    char *dir, *base, *magic, *buf;
+	    char *dir, *base, *magic;
 	    DIR *dirp;
 	    struct dirent *dp;
 	    int recursive = 0;
@@ -942,7 +981,8 @@ glob_helper(path, sub, flags, func, arg)
 
 	    magic = extract_elem(p);
 	    if (stat(dir, &st) < 0) {
-	        if (errno != ENOENT) rb_sys_warning(dir);
+	        if (errno != ENOENT)
+		    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)dir, 0);
 	        free(base);
 	        free(magic);
 	        break;
@@ -951,15 +991,14 @@ glob_helper(path, sub, flags, func, arg)
 		if (m && strcmp(magic, "**") == 0) {
 		    int n = strlen(base);
 		    recursive = 1;
-		    buf = ALLOC_N(char, n+strlen(m)+3);
+		    REALLOC_N(buf, char, n+strlen(m)+3);
 		    sprintf(buf, "%s%s", base, *base ? m : m+1);
 		    status = glob_helper(buf, buf+n, flags, func, arg);
-		    free(buf);
 		    if (status) goto finalize;
 		}
 		dirp = opendir(dir);
 		if (dirp == NULL) {
-		    rb_sys_warning(dir);
+		    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)dir, 0);
 		    free(base);
 		    free(magic);
 		    break;
@@ -983,11 +1022,11 @@ glob_helper(path, sub, flags, func, arg)
 			continue;
 		    if (fnmatch("*", dp->d_name, flags) != 0)
 			continue;
-		    buf = ALLOC_N(char, strlen(base)+NAMLEN(dp)+strlen(m)+6);
+		    REALLOC_N(buf, char, strlen(base)+NAMLEN(dp)+strlen(m)+6);
 		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
 		    if (lstat(buf, &st) < 0) {
-			if (errno != ENOENT) rb_sys_warning(buf);
-			free(buf);
+			if (errno != ENOENT)
+			    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)buf, 0);
 			continue;
 		    }
 		    if (S_ISDIR(st.st_mode)) {
@@ -995,24 +1034,22 @@ glob_helper(path, sub, flags, func, arg)
 		        strcpy(t, "/**");
 			strcpy(t+3, m);
 			status = glob_helper(buf, t, flags, func, arg);
-			free(buf);
 			if (status) break;
 			continue;
 		    }
-		    free(buf);
 		    continue;
 		}
 		if (fnmatch(magic, dp->d_name, flags) == 0) {
-		    buf = ALLOC_N(char, strlen(base)+NAMLEN(dp)+2);
+		    REALLOC_N(buf, char, strlen(base)+NAMLEN(dp)+2);
 		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
 		    if (!m) {
 			status = glob_call_func(func, buf, arg);
-			free(buf);
 			if (status) break;
 			continue;
 		    }
 		    tmp = ALLOC(struct d_link);
 		    tmp->path = buf;
+		    buf = 0;
 		    *tail = tmp;
 		    tail = &tmp->next;
 		}
@@ -1029,15 +1066,14 @@ glob_helper(path, sub, flags, func, arg)
 			    if (S_ISDIR(st.st_mode)) {
 				int len = strlen(link->path);
 				int mlen = strlen(m);
-				char *t = ALLOC_N(char, len+mlen+1);
 
-				sprintf(t, "%s%s", link->path, m);
-				status = glob_helper(t, t+len, flags, func, arg);
-				free(t);
+				REALLOC_N(buf, char, len+mlen+1);
+				sprintf(buf, "%s%s", link->path, m);
+				status = glob_helper(buf, buf+len, flags, func, arg);
 			    }
 			}
 			else {
-			    rb_sys_warning(link->path);
+			    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)link->path, 0);
 			}
 		    }
 		    tmp = link;
@@ -1050,6 +1086,8 @@ glob_helper(path, sub, flags, func, arg)
 	}
 	p = m;
     }
+    xfree(buf);
+    xfree(newpath);
     return status;
 }
 
@@ -1057,26 +1095,40 @@ static int
 rb_glob2(path, flags, func, arg)
     const char *path;
     int flags;
-    void (*func) _((const char*, VALUE));
+    void (*func) _((const char *, VALUE));
     VALUE arg;
 {
-    char *buf;
     int status;
 
-    buf = ALLOC_N(char, strlen(path)+1);
-    strcpy(buf, path);
-    status = glob_helper(buf, 0, flags, func, arg);
-    free(buf);
+    status = glob_helper(path, 0, flags, func, arg);
     return status;
 }
 
+struct rb_glob_args {
+    void (*func) _((const char*, VALUE));
+    VALUE arg;
+};
+
+static void
+rb_glob_caller(path, a)
+    VALUE path, a;
+{
+    struct rb_glob_args *args = (struct rb_glob_args *)a;
+    (*args->func)(RSTRING(path)->ptr, args->arg);
+}
+
 void
 rb_glob(path, func, arg)
     const char *path;
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    int status = rb_glob2(path, 0, func, arg);
+    struct rb_glob_args args;
+    int status;
+
+    args.func = func;
+    args.arg = arg;
+    status = rb_glob2(path, 0, rb_glob_caller, (VALUE)&args);
 
     if (status) rb_jump_tag(status);
 }
@@ -1087,7 +1139,12 @@ rb_globi(path, func, arg)
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    int status = rb_glob2(path, FNM_CASEFOLD, func, arg);
+    struct rb_glob_args args;
+    int status;
+
+    args.func = func;
+    args.arg = arg;
+    status = rb_glob2(path, FNM_CASEFOLD, rb_glob_caller, (VALUE)&args);
 
     if (status) rb_jump_tag(status);
 }
@@ -1097,14 +1154,7 @@ push_pattern(path, ary)
     const char *path;
     VALUE ary;
 {
-    VALUE str = rb_tainted_str_new2(path);
-
-    if (ary) {
-	rb_ary_push(ary, str);
-    }
-    else {
-	rb_yield(str);
-    }
+    rb_ary_push(ary, rb_tainted_str_new2(path));
 }
 
 static int
@@ -1117,18 +1167,19 @@ push_globs(ary, s, flags)
 }
 
 static int
-push_braces(ary, s, flags)
+push_braces(ary, str, flags)
     VALUE ary;
-    const char *s;
+    const char *str;
     int flags;
 {
-    char *buf, *b;
-    const char *p, *t;
+    char *buf = 0;
+    char *b;
+    const char *s, *p, *t;
     const char *lbrace, *rbrace;
     int nest = 0;
     int status = 0;
 
-    p = s;
+    s = p = str;
     lbrace = rbrace = 0;
     while (*p) {
 	if (*p == '{') {
@@ -1148,26 +1199,31 @@ push_braces(ary, s, flags)
 
     if (lbrace && rbrace) {
 	int len = strlen(s);
-	buf = xmalloc(len + 1);
-	memcpy(buf, s, lbrace-s);
-	b = buf + (lbrace-s);
 	p = lbrace;
 	while (*p != '}') {
 	    t = p + 1;
 	    for (p = t; *p!='}' && *p!=','; p++) {
 		/* skip inner braces */
-		if (*p == '{') while (*p!='}') p++;
+		if (*p == '{') {
+		    nest = 1;
+		    while (*++p != '}' || --nest) {
+			if (*p == '{') nest++;
+		    }
+		}
 	    }
+	    REALLOC_N(buf, char, len+1);
+	    memcpy(buf, s, lbrace-s);
+	    b = buf + (lbrace-s);
 	    memcpy(b, t, p-t);
 	    strcpy(b+(p-t), rbrace+1);
 	    status = push_braces(ary, buf, flags);
 	    if (status) break;
 	}
-	free(buf);
     }
     else {
-	status = push_globs(ary, s, flags);
+	status = push_globs(ary, str, flags);
     }
+    xfree(buf);
 
     return status;
 }
@@ -1179,39 +1235,29 @@ rb_push_glob(str, flags)
     VALUE str;
     int flags;
 {
-    const char *p, *pend;
-    char *buf;
-    char *t;
+    const char *p, *pend, *buf;
     int nest, maxnest;
     int status = 0;
     int noescape = flags & FNM_NOESCAPE;
     VALUE ary;
 
-    if (rb_block_given_p())
-	ary = 0;
-    else
-	ary = rb_ary_new();
-
+    ary = rb_ary_new();
     SafeStringValue(str);
-    buf = xmalloc(RSTRING(str)->len + 1);
-
     p = RSTRING(str)->ptr;
     pend = p + RSTRING(str)->len;
 
     while (p < pend) {
-	t = buf;
 	nest = maxnest = 0;
 	while (p < pend && isdelim(*p)) p++;
+	buf = p;
 	while (p < pend && !isdelim(*p)) {
 	    if (*p == '{') nest++, maxnest++;
 	    if (*p == '}') nest--;
 	    if (!noescape && *p == '\\') {
-		*t++ = *p++;
-		if (p == pend) break;
+		if (++p == pend) break;
 	    }
-	    *t++ = *p++;
+	    p++;
 	}
-	*t = '\0';
 	if (maxnest == 0) {
 	    status = push_globs(ary, buf, flags);
 	    if (status) break;
@@ -1222,10 +1268,11 @@ rb_push_glob(str, flags)
 	}
 	/* else unmatched braces */
     }
-    free(buf);
-
     if (status) rb_jump_tag(status);
-
+    if (rb_block_given_p()) {
+	rb_ary_each(ary);
+	return Qnil;
+    }
     return ary;
 }
 
@@ -1247,14 +1294,38 @@ dir_s_aref(obj, str)
 /*
  *  call-seq:
  *     Dir.glob( string, [flags] ) => array
- *     Dir.glob( string, [flags] ) {| filename | block }  => false
+ *     Dir.glob( string, [flags] ) {| filename | block }  => nil
  *
  *  Returns the filenames found by expanding the pattern given in
  *  <i>string</i>, either as an <i>array</i> or as parameters to the
  *  block. Note that this pattern is not a regexp (it's closer to a
- *  shell glob). See <code>File::fnmatch</code> for
- *  details of file name matching and the meaning of the <i>flags</i>
- *  parameter.
+ *  shell glob). See <code>File::fnmatch</code> for the meaning of
+ *  the <i>flags</i> parameter.
+ *
+ *  <code>*</code>::        Matches any file. Can be restricted by
+ *                          other values in the glob. <code>*</code>
+ *                          will match all files; <code>c*</code> will
+ *                          match all files beginning with
+ *                          <code>c</code>; <code>*c</code> will match
+ *                          all files ending with <code>c</code>; and
+ *                          <code>*c*</code> will match all files that
+ *                          have <code>c</code> in them (including at
+ *                          the beginning or end). Equivalent to
+ *                          <code>/ .* /x</code> in regexp.
+ *  <code>**</code>::       Matches directories recursively.
+ *  <code>?</code>::        Matches any one character. Equivalent to
+ *                          <code>/.{1}/</code> in regexp.
+ *  <code>[set]</code>::    Matches any one character in +set+.
+ *                          Behaves exactly like character sets in
+ *                          Regexp, including set negation
+ *                          (<code>[^a-z]</code>).
+ *  <code>{p,q}</code>::    Matches either literal <code>p</code> or
+ *                          literal <code>q</code>. Matching literals
+ *                          may be more than one character in length.
+ *                          More than two literals may be specified.
+ *                          Equivalent to pattern alternation in
+ *                          regexp.
+ *  <code>\</code>::        Escapes the next metacharacter.
  *
  *     Dir["config.?"]                     #=> ["config.h"]
  *     Dir.glob("config.?")                #=> ["config.h"]
@@ -1264,6 +1335,19 @@ dir_s_aref(obj, str)
  *     Dir.glob("*")                       #=> ["config.h", "main.rb"]
  *     Dir.glob("*", File::FNM_DOTMATCH)   #=> [".", "..", "config.h", "main.rb"]
  *
+ *     rbfiles = File.join("**", "*.rb")
+ *     Dir.glob(rbfiles)                   #=> ["main.rb",
+ *                                              "lib/song.rb",
+ *                                              "lib/song/karaoke.rb"]
+ *     libdirs = File.join("**", "lib")
+ *     Dir.glob(libdirs)                   #=> ["lib"]
+ *
+ *     librbfiles = File.join("**", "lib", "**", "*.rb")
+ *     Dir.glob(librbfiles)                #=> ["lib/song.rb",
+ *                                              "lib/song/karaoke.rb"]
+ *
+ *     librbfiles = File.join("**", "lib", "*.rb")
+ *     Dir.glob(librbfiles)                #=> ["lib/song.rb"]
  */
 static VALUE
 dir_s_glob(argc, argv, obj)
@@ -1282,6 +1366,21 @@ dir_s_glob(argc, argv, obj)
     return rb_push_glob(str, flags);
 }
 
+static VALUE
+dir_open_dir(path)
+    VALUE path;
+{
+    VALUE dir = rb_funcall(rb_cDir, rb_intern("open"), 1, path);
+
+    if (TYPE(dir) != T_DATA ||
+	RDATA(dir)->dfree != (RUBY_DATA_FUNC)free_dir) {
+	rb_raise(rb_eTypeError, "wrong argument type %s (expected Dir)",
+		 rb_obj_classname(dir));
+    }
+    return dir;
+}
+
+
 /*
  *  call-seq:
  *     Dir.foreach( dirname ) {| filename | block }  => nil
@@ -1305,7 +1404,7 @@ dir_foreach(io, dirname)
 {
     VALUE dir;
 
-    dir = rb_funcall(rb_cDir, rb_intern("open"), 1, dirname);
+    dir = dir_open_dir(dirname);
     rb_ensure(dir_each, dir, dir_close, dir);
     return Qnil;
 }
@@ -1327,7 +1426,7 @@ dir_entries(io, dirname)
 {
     VALUE dir;
 
-    dir = rb_funcall(rb_cDir, rb_intern("open"), 1, dirname);
+    dir = dir_open_dir(dirname);
     return rb_ensure(rb_Array, dir, dir_close, dir);
 }
 
@@ -1341,8 +1440,27 @@ dir_entries(io, dirname)
  *  similar to shell filename globbing. It may contain the following
  *  metacharacters:
  *
- *  <i>flags</i> is a bitwise OR of the <code>FNM_xxx</code> parameters.
- *  The same glob pattern and flags are used by <code>Dir::glob</code>.
+ *  <code>*</code>::        Matches any file. Can be restricted by
+ *                          other values in the glob. <code>*</code>
+ *                          will match all files; <code>c*</code> will
+ *                          match all files beginning with
+ *                          <code>c</code>; <code>*c</code> will match
+ *                          all files ending with <code>c</code>; and
+ *                          <code>*c*</code> will match all files that
+ *                          have <code>c</code> in them (including at
+ *                          the beginning or end). Equivalent to
+ *                          <code>/ .* /x</code> in regexp.
+ *  <code>?</code>::        Matches any one character. Equivalent to
+ *                          <code>/.{1}/</code> in regexp.
+ *  <code>[set]</code>::    Matches any one character in +set+.
+ *                          Behaves exactly like character sets in
+ *                          Regexp, including set negation
+ *                          (<code>[^a-z]</code>).
+ *  <code>\</code>::        Escapes the next metacharacter.
+ *
+ *  <i>flags</i> is a bitwise OR of the <code>FNM_xxx</code>
+ *  parameters. The same glob pattern and flags are used by
+ *  <code>Dir::glob</code>.
  *
  *     File.fnmatch('cat',       'cat')        #=> true
  *     File.fnmatch('cat',       'category')   #=> false
