@@ -2,8 +2,8 @@
 
   sprintf.c -
 
-  $Author: nobu $
-  $Date: 2004/07/17 23:14:52 $
+  $Author: matz $
+  $Date: 2005/09/09 13:15:15 $
   created at: Fri Oct 15 10:39:26 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -107,7 +107,7 @@ sign_bits(base, p)
 	       (posarg = -1, GETNTHARG(n))))
 
 #define GETNTHARG(nth) \
-    ((nth >= argc) ? (rb_raise(rb_eArgError, "too few arguments."), 0) : argv[nth])
+    ((nth >= argc) ? (rb_raise(rb_eArgError, "too few arguments"), 0) : argv[nth])
 
 #define GETASTER(val) do { \
     t = p++; \
@@ -249,6 +249,7 @@ rb_f_sprintf(argc, argv)
     fmt = GETNTHARG(0);
     if (OBJ_TAINTED(fmt)) tainted = 1;
     StringValue(fmt);
+    fmt = rb_str_new4(fmt);
     p = RSTRING(fmt)->ptr;
     end = p + RSTRING(fmt)->len;
     blen = 0;
@@ -770,14 +771,13 @@ rb_f_sprintf(argc, argv)
     }
 
   sprint_exit:
-#if 0
-    /* XXX - We cannot validiate the number of arguments because
-     *       the format string may contain `n$'-style argument selector.
+    /* XXX - We cannot validiate the number of arguments if (digit)$ style used.
      */
-    if (RTEST(ruby_verbose) && nextarg < argc) {
-	rb_raise(rb_eArgError, "too many arguments for format string");
+    if (posarg >= 0 && nextarg < argc) {
+	const char *mesg = "too many arguments for format string";
+	if (RTEST(ruby_debug)) rb_raise(rb_eArgError, mesg);
+	if (RTEST(ruby_verbose)) rb_warn(mesg);
     }
-#endif
     rb_str_resize(result, blen);
 
     if (tainted) OBJ_TAINT(result);
